#include "KDTree.hxx"

//------------------------------------------------------------------------------------
bool KDTree::Intersect(Ray& ray)
{
    // intersect the ray with the bounding box of the kdtree
    std::pair<float, float> tmm = boundingBox.Intersect(ray);

    // check if intersection is valid
    if (tmm.first >= FLT_MAX - Epsilon) return false;

    // start traversal from the root node
    return Intersect(rootNode, ray, tmm.first, tmm.second);
}

//------------------------------------------------------------------------------------
bool KDTree::Intersect(Node* node, Ray& ray, float min, float max)
{
    // don't do anything for null nodes
    if (node == NULL) return false;

    // if leaf node, then look for intersection with primitives
    if (node->isLeafNode())
    {
        return node->LeafIntersect(ray, min, max);
    }
    
    // get near and far child
    Node* near = NULL;
    Node* far = NULL;
    node->GetNearFar(ray, near, far);

    // compute distance to the split plane
    float dist = node->DistanceToSplitPlane(ray);

    // the whole interval is on near side
    if (dist < 0 || dist > max)
    {
        return Intersect(near, ray, min, max);
        
    // whole interval is on far side
    }else if (dist <= min)
    {
        return Intersect(far, ray, min, max);

    // the interval intersects the plane
    }else
    {
        // first test near side
        if (Intersect(near, ray, min, dist)) return true;

        // then test the far side
        return Intersect(far, ray, dist, max);
    }
    
    // we shouldn't get upto here, if yes then this is a bug
    assert(0);
    return false;
}


//------------------------------------------------------------------------------------
bool KDTree::Node::LeafIntersect(Ray& ray, float min, float max)
{
    // here we store a copy of the ray
    Ray r = ray;
    bool found = false;
        
    // search for the nearest intersection
    for (unsigned int i=0; i < primitive.size(); i++)
    {
        // if there is an intersection
        if (primitive[i]->Intersect(r))
        {
            // do only proceed if the intersection is near than previous
            if (r.t >= min - Epsilon && r.t < max)
            {
                ray = r;
                ray.hit = primitive[i];
                
                max = r.t;
                found = true;
            }
        }
    }
    
    return found;
}


//------------------------------------------------------------------------------------
void KDTree::BuildTree(const std::vector<Primitive*>& list, const Box& bbox, int count, int depth)
{
    // first deleter root node, so that the tree is rebuild
    if (rootNode) delete rootNode;

    // debug
    //std::cout << "Start building kd-tree (maxDepth = " << depth << ")" << std::endl;
    
    // create empty root node
    rootNode = new Node();
    
    // this number declares the maximal possible number of primitives in a leaf node
    maxPrimitives = count;
    maxDepth = depth;
    
    // reset the bounding box
    boundingBox = bbox;

    // copy primitives
    rootNode->primitive = list;

    // create the tree recursively, the children start as y-axis
    BuildTree(rootNode, X_AXIS, 0);

    // debug
    //std::cout << "Building kd-tree done" << std::endl;
}

//------------------------------------------------------------------------------------
void KDTree::BuildTree(Node* node, Axis axis, int depth)
{
    // just for debug
    assert(node != NULL);
    assert(node->left == NULL);
    assert(node->right == NULL);

    // setup node's split axis and plane
    node->splitAxis = axis;
    node->splitPlane = FindSplitPlane(node);
    
    // do only subdivide current node if primitive number is still over the maximum
    if (node->primitive.size() > maxPrimitives && (unsigned int)depth < maxDepth )
    {
        // create both childrens
        node->left = new Node();
        node->right = new Node();

        // check based on the split plane where to put the primitives
        for (unsigned int i=0; i < node->primitive.size(); i++)
        {
            const Box& box = node->primitive[i]->GetBounds();

            // if primitive lies completly on the right side
            if (box.min[axis] >= node->splitPlane)
            {
                node->right->primitive.push_back(node->primitive[i]);
            
            // if primitive lies completly on the left side
            }else if (box.max[axis] <= node->splitPlane)
            {
                node->left->primitive.push_back(node->primitive[i]);

            // for the rest of cases we just put the primitive in both subtrees
            }else
            {
                node->left->primitive.push_back(node->primitive[i]);
                node->right->primitive.push_back(node->primitive[i]);
            }
        }

        // clear the node's primitive list, since now the childrens do contain them
        node->primitive.clear();
                
        // setup the subtrees
        BuildTree(node->left, Axis(((int)axis + 1) % 3), depth + 1);
        BuildTree(node->right, Axis(((int)axis + 1) % 3), depth + 1);
    
        // do return
        return;
    }

    // we can only be here, if we have reached a leaf node
    node->splitAxis = NO_AXIS;
}

//------------------------------------------------------------------------------------
float KDTree::FindSplitPlane(Node* node)
{
    if (node == NULL) return Infinity;

    // temporary variable
    float avg = 0.0;
    
    // to find a location for the split plane we use the method of arithmetic average value.
    // Hence we compute the average value of the prmitive's location
    // and set this as the split plane
    // This is a bad heuristic. Optimize it later if you like
    for (unsigned int i=0; i < node->primitive.size(); i++)
    {
        avg += (node->primitive[i]->GetBounds().min[node->splitAxis] + node->primitive[i]->GetBounds().max[node->splitAxis]) * 0.5;
    }

    avg /= float(node->primitive.size());

    return avg;
}



