#ifndef KDTREE_H
#define KDTREE_H

#include "Box.hxx"
#include "Primitive.hxx"

class KDTree
{
private:
    //! Axis id
    enum Axis
    {
        X_AXIS = 0,
        Y_AXIS = 1,
        Z_AXIS = 2,
        NO_AXIS= 3
    };

    //! Subclass representing a node in the tree
    class Node
    {
        friend class KDTree;
        private:
            Axis  splitAxis;
            float splitPlane;
            
            Node* left;
            Node* right;

            std::vector<Primitive*> primitive;

            //! Setup default values
            Node()
            {
                left = right = NULL;
                splitAxis = NO_AXIS;
                splitPlane = 0.0;
            }

            //! Release used memory
            ~Node()
            {
                if (left) delete left;
                if (right) delete right;
            }

            //! Check whenever the node is a leaf node
            inline bool isLeafNode() const { return splitAxis == NO_AXIS; }

            /**
             * Check if ray intersects any object in this node (leaf node)
             * Return true if intersection is found and store the intersection
             * parameters (t and hit) in the ray structure
             **/
            bool LeafIntersect(Ray& ray, float min, float max);

            /**
             * Returns the children nodes sorted as near and far.
             * Near child is the one which contains the ray origin
             **/
            inline bool GetNearFar(const Ray& r, Node*& near, Node*& far)
            {
                if (splitPlane >= r.org[splitAxis])
                {
                    near = left;
                    far = right;
                }else
                {
                    near = right;
                    far = left;
                }
                return true;
            }

            //! Compute distance to the splitting plane
            inline float DistanceToSplitPlane(const Ray& ray) const
            {
                 return (splitPlane - ray.org[splitAxis]) / ray.dir[splitAxis];
            }
            
    };
    

    //! Maximal number of primitives in a leaf node
    unsigned int maxPrimitives;

    //! Tree depth which should be reached
    unsigned int maxDepth;
    
    //! Root node of the tree
    Node* rootNode;

public:

    //! bounding box of the whole tree
    Box boundingBox;
    
    /**
     * Intersect a ray with the kdtree.
     * Return true if intersection is found. Store corresponding
     * intersection information if the ray structure (t, hit)
     **/
    bool Intersect(Ray& ray);

    /**
     * Build the kd-tree from the given primitive list.
     **/
    void BuildTree(const std::vector<Primitive*>& list, const Box& bbox, int maxPrimitives = 4, int maxDepth = 10);
    
    
    //! Initialize
    KDTree()
    {
        rootNode = NULL;
        maxPrimitives = 4;
    }
    
    //! Cleanup used memory
    ~KDTree()
    {
        if (rootNode) delete rootNode;
    }

private:

    //! Recursive method to look for the intersection
    bool Intersect(Node* node, Ray& ray, float min, float max);

    //! Recursive tree building method
    void BuildTree(Node* node, Axis axis, int depth);
    
    //! Find optimal split location
    float FindSplitPlane(Node* node);
    
};


#endif
