#ifndef LIGHFIELD_H
#define LIGHFIELD_H


#include "Texture2D.hxx"
#include "PerspectiveCamera.hxx"
#include "InfinitePlane.hxx"
#include "Scene.hxx"

class Lightfield
{

public:

    // sample structure to hold the lightfiled
    typedef struct _Sample
    {
        // Texture containing the lf sample
        Texture2D* texture;

        // Camera used to sample this image
        PerspectiveCamera* camera;

    }Sample;



    Lightfield()
    {
        mImage = NULL;
        resX = 0;
        resY = 0;
    }

    virtual ~Lightfield()
    {
        for (unsigned int i=0; i < getSamplesNum(); i++)
        {
            delete getSample(i).texture;
            delete getSample(i).camera;
        }
    }
    

    // Get nubmer of samples
    inline unsigned int getSamplesNum() const { return mSample.size(); }

    // Get sample 
    inline Sample& getSample(unsigned int i) { return mSample[i]; }

    // add sample 
    inline void addSample(const Sample& s) { mSample.push_back(s); }

    // set resolution of the final image 
    inline void setResolution(int w, int h) { resX = w; resY = h; }

    /**
     * Refocus the image samples and store result into the resulting image.
     * @param focalPlane new position of the focal plane.
    */
    void refocus(float focalPlane)
    {
        // we must have at least one sample 
        if (getSamplesNum() == 0) return;
        if (resX == 0) resX = mVirtualCamera->resX;
        if (resY == 0) resY = mVirtualCamera->resY;

        // generate resulting image
        if (mImage) delete mImage;
        mImage = new Image(resX, resY);

        // for each pixel in the resulting image do
        for (int y=0; y < mImage->getHeight(); y++)
        for (int x=0; x < mImage->getWidth(); x++)
        {
            Vec4f& pixel = mImage->getPixel(x,y);

            // ------------ IMPLEMENT HERE ----------------------


            // for each sample camera do now
            for (unsigned int i=0; i < getSamplesNum(); i++)
            {
                // ------------ IMPLEMENT HERE ----------------------
            }
        }
    }


    // get resulting image
    inline Image* getResultImage() { return mImage; }

    // get current virtual camera
    inline PerspectiveCamera* getVirtualCamera() { return mVirtualCamera; }


    /**
    * Sample lightfield for the given scene.
    * @param uNumber Number of samples in x direction
    * @param vNumber Number of samples in y direction
    * @param uDim Distance how far to move the sampling camera in x direction
    * @param vDim same as uDim but for y direction
    * @param centerCamera Camera which is in the center of all the samples.
    * @param scene Scene to sample from
    * @param fileName Name of the file used to store the samples (must containd 2x %d)
    **/
    void sampleLightfield(int uNumber, int vNumber, float uDim, float vDim, PerspectiveCamera& centerCamera, Scene& scene, char *fileName)
    {
        mVirtualCamera = &centerCamera;

        // for each image in the field do
        for (int v=0; v < vNumber; v++)
        for (int u=0; u < uNumber; u++)
        {
            // prepare filename
            char file[256];
            sprintf(file, fileName, u,v);

            // setup image
            Image*  img = new Image(centerCamera.resX, centerCamera.resY);

            // create camera for this frame
            PerspectiveCamera* camera = new PerspectiveCamera(centerCamera);
                
            // now based on the wished number of the samples of the light field
            // in the corresponding dimension we setup the step width to move the
            // virtual cameras away from the center one
            float uDist = 2.0 * uDim / (float)uNumber;
            float vDist = 2.0 * vDim / (float)vNumber;
            
            // compute new coordinate of the camera origin
            Vec3f pos = camera->getPos();
            float du = (u - uNumber/2) * uDist;
            float dv = (v - vNumber/2) * vDist;
            pos = pos + camera->getXAxis() * du + camera->getYAxis() * dv;
            camera->setPos(pos);
            
            // since we move our sampling cameras around the center one, we have
            // to compute a shearing component to make sure that the image plane of
            // all the camera's samples lies on the same position.

            // ------------ IMPLEMENT HERE ----------------------

            // check if image already exists, then just load it
            FILE* f = fopen(file, "rb");
            if (f)
            {
                fclose(f);
                img->LoadImage(file);

            // file does not exists, hence render it
            }else
            {                            
                // just some debug info
                printf("render %s (%dx%d)\n", file, camera->resX, camera->resY);
    
                // now render the scene
                for (int y=0; y < camera->resY; y++)
                    for (int x=0; x < camera->resX; x++)
                    {
                        // primary ray to trace
                        Ray ray;
    
                        // initialize the ray through the new sample
                        camera->InitRay(x, y, ray);
    
                        // ray trace the scene
                        img->setPixel(Min(Max(scene.RayTrace(ray),Vec3f(0)),Vec3f(1)), x, y);
                    }
    
                // write final image
                img->WritePNG(file);
            }

            // store this into the lightfield
            Lightfield::Sample sample;
            sample.camera = camera;
            sample.texture = new Texture2D(img);
            sample.texture->setBorderColor(Vec4f(0,0,0,0));
            sample.texture->setWrapMode(Texture::CLAMP_TO_BORDER);
            //sample.texture->setInterpolationMode(Texture::LINEAR);
            addSample(sample);

        }
    }

private:

    // samples
    std::vector<Sample> mSample;

    // resulting image generated by the lightfield
    Image* mImage;

    // resolution of the final image 
    int resX, resY;

    // virtual camera
    PerspectiveCamera* mVirtualCamera;

};


#endif


