
#include "Image.hxx"
#include "PerspectiveCamera.hxx"
#include "Object.hxx"
#include "Scene.hxx"

#include "PhongShader.hxx"
#include "PointLight.hxx"
#include "FlatShader.hxx"
#include "FlatTransparentShader.hxx"
#include "CloudShader.hxx"

#include "Lightfield.hxx"


// --------------------------------------------------------------------------
Object* AddObject(char* fileName, Scene& scene, bool useAcc = true, const Object::PrimitiveFactory& factory = Object::TriangleFactory() )
{
    Object* obj = new Object(useAcc);
    obj->ParseOBJ(fileName, factory);
    obj->setColor(Vec3f(1,1,1));
    obj->BuildAccelStructure();

    scene.AddObject(obj);

    return obj;
}


// --------------------------------------------------------------------------
void RenderFrame(Camera& camera, Scene& scene, char *fileName)
{    
    // just some debug info
    printf("render %s (%dx%d)\n", fileName, camera.resX, camera.resY);

    // setup image and primary ray
    Image img(camera.resX, camera.resY);

    // now render the scene
    for (int y=0; y < camera.resY; y++)
        for (int x=0; x < camera.resX; x++)
        {
            // primary ray to trace
            Ray ray;

            // initialize the ray through the new sample
            camera.InitRay(x, y, ray);

            // ray trace the scene 
            img.setPixel(Min(Max(scene.RayTrace(ray),Vec3f(0)),Vec3f(1)), x, y);
        }

    // write final image
    img.WritePNG(fileName);
}


// --------------------------------------------------------------------------
#define RESX 256 // image x-resolution
#define RESY 256 // image y-resolution
#define USE_KDTREE 1

// --------------------------------------------------------------------------
int main(int argc, char **argv)
{
    // setup camera
    PerspectiveCamera camera(Vec3f(0.0, 0.0, -0.225),Vec3f(0,0,1),Vec3f(0,1,0), 60, RESX, RESY);

    // setup default scene 
    Scene* scene = new Scene();
    scene->setBgColor(Vec3f(0.0,0.0,0.0));

    // setup lights
    PointLight* light1 = new PointLight(Vec3f(0.1, 5.03, -10.2), Vec3f(150.0));
    scene->AddLight(light1);
    scene->castShadows = false;

    // setup cow 
    Object* sphere = AddObject("mesh/sphere.obj", *scene, USE_KDTREE, Object::TexturedSmoothTriangleFactory());
    Shader* phong = new PhongShader(scene, Vec3f(0.8, 0.7, 0.6), 0, 0.8, 0.4, 12.0);
    Texture* cowtex = new Texture2D("textures/cowspots.png");
    phong->addTexture(cowtex);
    sphere->setShader(phong);

    // create light field for the scene with certain amount of samples
    Lightfield* lightfield = new Lightfield();
    lightfield->setResolution(256, 256);
    lightfield->sampleLightfield(16, 16, 0.02f, 0.02f, camera, *scene, "results/%d_%d.png");

    float focal = camera.getFocalPlane();
    printf("cam focal %f\n", focal);

    // refocus the lightfield and write results into a file
    for (float f=0.0; f < focal; f += focal / 20.0)
    {
        char file[128];
        sprintf(file, "lf_%f.png", f);
        printf("render %s\n", file);


        lightfield->refocus(f);
        lightfield->getResultImage()->WritePNG(file);
    }

    delete lightfield;

    // delete used data
    delete cowtex;
    delete phong;


    delete scene;
}
