#ifndef OBJECT_HXX
#define OBJECT_HXX

#include "TexturedSmoothTriangle.hxx"
#include "SmoothTriangle.hxx"
#include "KDTree.hxx"
#include "Shader.hxx"

class Object
{
    std::vector<Primitive *> primitive; // primitives

    // add a new primitive
    void Add(Primitive *prim)
    {
        primitive.push_back(prim);
    };

    KDTree* kdTree;
    bool m_useAcc;  // flag if we have to use the acceleration structure
    Box boundingBox;
    Vec3f color;

    bool m_castShadows;

public:

    //! factory which is able to generate primitives
    class PrimitiveFactory
    {
    public:
        virtual ~PrimitiveFactory(){}
        virtual Primitive* Create(Vec3f va, Vec3f vb, Vec3f vc, Vec3f na, Vec3f nb, Vec3f nc, Vec2f ta, Vec2f tb, Vec2f tc) const = 0;
    };

    //! primitive factory which will generate simple triangles
    class TriangleFactory : public PrimitiveFactory
    {
        Primitive* Create(Vec3f va, Vec3f vb, Vec3f vc, Vec3f na, Vec3f nb, Vec3f nc, Vec2f ta, Vec2f tb, Vec2f tc) const
        {
            return new Triangle(va, vb, vc);
        }
    };

    //! primitive factory which will generate smooth triangles
    class SmoothTriangleFactory : public PrimitiveFactory
    {
        Primitive* Create(Vec3f va, Vec3f vb, Vec3f vc, Vec3f na, Vec3f nb, Vec3f nc, Vec2f ta, Vec2f tb, Vec2f tc) const
        {
            return new SmoothTriangle(va, vb, vc, na, nb, nc);
        }
    };

    //! primitive factory which will generate textured smooth triangles
    class TexturedSmoothTriangleFactory : public PrimitiveFactory
    {
        Primitive* Create(Vec3f va, Vec3f vb, Vec3f vc, Vec3f na, Vec3f nb, Vec3f nc, Vec2f ta, Vec2f tb, Vec2f tc) const
        {
            return new TexturedSmoothTriangle(va, vb, vc, na, nb, nc, ta, tb, tc);
        }
    };

    //! Constructor
    Object(bool useAcc)
    {
        m_useAcc = useAcc;
        color = Vec3f(0,0,1);
        kdTree = NULL;
        m_castShadows = true;
    };

    ~Object() 
    {
        // cleanup primitives
        for (unsigned int i=0;i<primitive.size();i++)
            delete primitive[i];
                
        // delete tree
        if (kdTree) delete kdTree;
    };

    // get color 
    const Vec3f& getColor() const { return color; }

    // set color of the object 
    void setColor(const Vec3f& v) { color = v; }

    // set a shader to all the primitives of the object 
    void setShader(Shader* sh) 
    {
        for (unsigned int i=0;i<primitive.size();i++)
            primitive[i]->setShader(sh);
    }

    // parse obj file
    void ParseOBJ(char *fileName, const PrimitiveFactory& factory = TriangleFactory() );

    // intersect all contained objects
    bool Intersect(Ray &ray)
    {
        if(m_useAcc)
        {
            return kdTree->Intersect(ray);
        }
        else
        {
            bool hit = false;

            for (unsigned int i=0;i<primitive.size();i++)
                hit |= primitive[i]->Intersect(ray);
            return hit;
        }
        return false;
    };

    // find occluder
    bool Occluded(Ray &ray)
    {
        if(m_useAcc)
        {
            return kdTree->Intersect(ray);
        }
        else
        {
            for (unsigned int i=0;i<primitive.size();i++)
                if (primitive[i]->Occluded(ray))
                    return true;

            return false;
        }
        return false;
    };

    // get normal of the object for a certain ray
    Vec3f GetNormal(Ray &ray)
    {
        std::cerr << "Object::GetNormal is not defined!" << std::endl;
        return Vec3f(0,0,0); // dummy !
    };

    // calculate bounding box
    Box CalcBounds()
    {
        boundingBox.Clear();

        for (int i=0;i<(int)primitive.size();i++) 
        {
            Box box = primitive[i]->GetBounds();
            boundingBox.Extend(box);
        }

        return boundingBox;
    }

    // build acceleration structure to increase rendering performance
    void BuildAccelStructure()
    {
        if (!m_useAcc) return;

        // compute the bounding box
        boundingBox = CalcBounds();

        // recreate the tree if already created
        if (kdTree) delete kdTree;
        kdTree = new KDTree();

        // build acceleration data structure
        kdTree->BuildTree(primitive, boundingBox);
    }

    // does this object cast shadows
    inline bool isCastShadows() const { return m_castShadows; }
    inline void setCastShadows(bool b)
    {
        // this object's cast shadow flag
        m_castShadows = b;

        // set to each primitive 
        for (int i=0;i<(int)primitive.size();i++) 
        {
            primitive[i]->castShadows() = b;
        }
    }

};

#endif
