#ifndef PERSPECTIVECAMERA_HXX
#define PERSPECTIVECAMERA_HXX

#include "Camera.hxx"
#include "Matrix4f.hxx"


class PerspectiveCamera : public Camera
{
public:
	// input values
	Vec3f pos,dir,up;
	float fovV, fovH;

	// preprocessed values
	//float focus;
	Vec3f xAxis,yAxis,zAxis;
	float aspect;
    float focalPlane;
    
    // frustum planes
    float frustumLeft, frustumRight;
    float frustumTop, frustumBottom;


public:
	PerspectiveCamera(Vec3f pos,Vec3f _dir,Vec3f up,
		float fov,
		int resX, int resY
		)
		: Camera(resX,resY),pos(pos),dir(_dir),up(up),fovH(fov)
	{
        // normalize direction   
        Normalize(dir);
        Normalize(up);
        zAxis = dir;
        xAxis = Cross(dir,up);
        yAxis = Cross(xAxis,zAxis);

        Normalize(xAxis);
        Normalize(yAxis);
        Normalize(zAxis);

        // compute aspect ratio
        aspect = resX / float(resY);

        // compute focal plane and vertical fov
        focalPlane = fabs(1.0 / tan(2.0 * deg2rad(fovH) ));
        fovV = 2.0 * rad2deg(atan(aspect * tan(deg2rad(fovH) * 0.5)) );

        // setup symmetric frustum per default
        frustumLeft = -0.5;
        frustumRight = 0.5;
        frustumBottom = -0.5;
        frustumTop = 0.5;
	}

    //! Copy constructor
    PerspectiveCamera(const PerspectiveCamera& cam) :
        Camera(cam),
        pos(cam.pos),
        dir(cam.dir),
        up(cam.up),
        fovV(cam.fovV),
        fovH(cam.fovH),
        //focus(cam.focus),
        xAxis(cam.xAxis),
        yAxis(cam.yAxis),
        zAxis(cam.zAxis),
        aspect(cam.aspect),
        focalPlane(cam.focalPlane),
        frustumLeft(cam.frustumLeft),
        frustumRight(cam.frustumRight),
        frustumTop(cam.frustumTop),
        frustumBottom(cam.frustumBottom)
    {

    }

    virtual ~PerspectiveCamera()
    {}

    //! Init ray based on the frustum 
	virtual bool InitRay(float x, float y, Ray &ray)
	{
        ray.org = pos;
        ray.dir 
            = (  lerp(frustumLeft, frustumRight, x/(float)resX) * aspect * xAxis)
            + (  lerp(frustumTop, frustumBottom, y/(float)resY) * yAxis)
            + (focalPlane * zAxis);
        ray.t = Infinity;

        ray.hit = NULL;

        Normalize(ray.dir);

        return true;
	};

    /**
    * Setup viewing frustum. Use this method to specify a generic 
    * viewing frustum by its planes. Each plane represent a distance
    * from the camera origin on the focal plane.
    * It is better to have right - left <= 1 and top - bottom <= 1.
    **/
    inline void setFrustum(float left, float right, float bottom, float top)
    {
        frustumLeft = left;
        frustumRight = right;
        frustumTop = top;
        frustumBottom = bottom;   

        // since focal plane has changed we have to update the fovs
        fovH = 2.0 * rad2deg(atan(0.5 * (frustumRight - frustumLeft) * aspect / focalPlane));
        fovV = 2.0 * rad2deg(atan(0.5 * (frustumTop - frustumBottom) / focalPlane));
    }


    //! Get position
    inline const Vec3f& getPos() const { return pos; }

    //! Set new position
    inline void setPos(const Vec3f& p)
    {
        pos = p;
    }

    //! Return the distance to the focal plane (to the image plane    
    inline float getFocalPlane() const { return focalPlane; }

    //! Set distance to focal plane
    inline void setFocalPlane(float f)
    {
        focalPlane = f;

        // since focal plane has changed we have to update the fovs
        //fovH = 2.0 * rad2deg(atan(0.5 * (frustumRight - frustumLeft) * aspect / focalPlane));
        //fovV = 2.0 * rad2deg(atan(0.5 * (frustumTop - frustumBottom) / focalPlane));
    }

    inline Vec3f getXAxis() 
    {
        return xAxis;
    }    
    inline Vec3f getYAxis() 
    {
        return yAxis;
    }

    inline Vec3f getZAxis() 
    {
        return zAxis;
    }
    inline float getAspectRatio() const
    {
        return aspect;
    }
    inline float getVerticalFov() const
    {
        return fovV;
    }
    inline float getHorizontalFov() const
    {
        return fovH;
    }

};
#endif
