#ifndef AREALIGHT_HXX
#define AREALIGHT_HXX

#include "Light.hxx"

// quadrangular area light source

class QuadAreaLight : public Light
{
private:
    Vec3f power;
    Vec3f p0,e1,e2;
    float area;
    Vec3f normal;
    unsigned int raycount;
public:


    /**
     * Initialize quad area light.
     * @param power Intensity of the light source
     * @param pos Position of the light source
     * @param d1,d2 Direction vectors spanning the quad polygon
     **/
    QuadAreaLight(Vec3f power, Vec3f pos, Vec3f d1, Vec3f d2)
    : Light(),power(power),p0(pos),e1(d1),e2(d2), raycount(1)
    { 
        // compute normal by computing the cross product of both expansion vectors
        normal = Cross(e1,e2);

        // the area of a quad is the length of the cross product vector 
        area = Length(normal); 

        // we have always to normalize the normal
        Normalize(normal);
    };  

    //! Return number of shadow rays handled by this light source
    unsigned int GetNumberOfRays() const { return raycount; }

    //! Set number of rays
    void SetNumberOfRays(unsigned int i) { raycount = i; }

    /**
     * Illuminate the pixel/ray.
     * The illumination is done in the same as for a simple point light source,
     * but now we place the point light source randomly on the quad surface.
     * Also the area of the quad light source has to be taken into account, by
     * computing the intensity
     **/
    bool Illuminate(Ray &ray, Vec3f &intensity, unsigned int index = 0)
    {
        // compute random position of a point light on the quad
        float xi1 = frand();
        float xi2 = frand();
        Vec3f pos = p0 + xi1 * e1 + xi2 * e2;
                
        // the new ray direction is the direction to this point
        ray.dir = pos - ray.org;        
        ray.hit = NULL;
        ray.t = Length(ray.dir) - Epsilon;
        Normalize(ray.dir);        
        
        // compute attenuation factor 
        float att = 1.0f / (ray.t * ray.t);

        // sample quad light 
        float cosN = -Dot(ray.dir,normal);
        if (cosN < Epsilon) return false;
        
        // compute intensity based on the area of the light source
        intensity = att * area * cosN * power;
                
        return true;
    }

    //! Get normal of the quad representing the light source    
    const Vec3f& GetNormal() const
    {
        return normal;
    }

};

#endif
