#ifndef SCENE_HXX
#define SCENE_HXX

#include "Object.hxx"
#include "Light.hxx"

//! Class to encapsulate the whole scene (lights, objects, ...)
class Scene
{
public:

    //! List of objects in the scene
    std::vector<Object*> mObjects;

    //! List of all lights in the scene 
    std::vector<Light*> mLights;

    //! Background color
    Vec3f bgColor;

    //! Shall the scene cast the shadows
    bool castShadows;

    //! Constructor
    Scene() : castShadows(true)
    {

    }

    //! Destructor
    ~Scene()
    {
        // remove all objects
        for (unsigned int i=0; i < mObjects.size(); i++)
            if (mObjects[i]) delete mObjects[i];

        // remove all lights
        for (unsigned int i=0; i < mLights.size(); i++)
            if (mLights[i]) delete mLights[i];
    }

    //! Add a new object to the scene (Note: Scene will be the new owner!)
    void AddObject(Object* obj)
    {
        mObjects.push_back(obj);
    }

    //! Remove object
    void RemoveObject(Object* obj)
    {
        for (std::vector<Object*>::iterator it=mObjects.begin(); it != mObjects.end(); it++)
            if ((*it) == obj){
                delete (*it);
                mObjects.erase(it);
                break;
            }
    }

    //! Add a new light to the scene (Note: Scene will be the owner)
    void AddLight(Light* light)
    {
        light->mScene = this;
        mLights.push_back(light);
    }

    //! Remove light 
    void RemoveLight(Light* light)
    {
        for (std::vector<Light*>::iterator it=mLights.begin(); it != mLights.end(); it++)
            if ((*it) == light){
                delete (*it);
                mLights.erase(it);
                break;
            }
    }
    
    //! Set new background color
    void setBgColor(const Vec3f& c) { bgColor = c; }

    //! Get bg color
    const Vec3f& getBgColor() const { return bgColor; }

    /**
      * Raytrace the scene.
      * Return the shaded color of the ray (pixel) or the bgColor if ray 
      * do not intersect any object in the scene.
      * @param ray Ray to trace through the scene
      **/
    Vec3f RayTrace(Ray& ray)
    {
        // we use this value to manage maximal possible tracing depth
        static int maxBounce = 3;

        // here we store the result
        Vec3f result = bgColor;

        // only shoot another ray if the maximum recursion depth is not exceeded
        if (maxBounce > 0 && Intersect(ray))
        {
            maxBounce --;
            if (ray.hit->getShader())
                result = ray.hit->getShader()->Shade(ray); // shade the primitive
            else
                result = Vec3f(1,0,1);
            maxBounce ++;
        }

        return result;
    }


    //! intersect the ray within each object in the scene
    bool Intersect(Ray& ray)
    {
        bool hit = false;
        for (unsigned int i=0; i < mObjects.size(); i++)
        {
            hit |= mObjects[i]->Intersect(ray);
        }
        return hit;
    }

private:



};

#endif

