#ifndef TEXTURE2D_HXX
#define TEXTURE2D_HXX

#include "Texture.hxx"

class Texture2D : public Texture
{
private:
    //! image class to hold the data
    Image* mImage;

    //! Get texel, but use integer coordinates to sample it. Take into accoutn used wrap mode
    Vec4f _GetTexel(int px, int py)
    {
        if (mImage == NULL) return Vec4f();

        // compute correct texture coordiantes based on the wrap mode 
        switch (mMode)
        {
            case CLAMP:
            case CLAMP_TO_BORDER:
                // clamp coordinates
                px = max(0, min(px, resX - 1));
                py = max(0, min(py, resY - 1));
                break;

            case REPEAT:
                // repeat coordinates
                px = (resX + px % resX) % resX;
                py = (resY + py % resY) % resY;

                break;

            default:
                break;
        };

        // if clamp to border color
        if (mMode == CLAMP_TO_BORDER)
        {
            if (px == resX - 1 || px == 0 || py == 0 || py == resY - 1)
                return mBorderColor;
        }


        return mImage->getPixel(px, py);
    }

    //! Return interpolated texel value
    Vec4f _GetSmoothTexel(float u, float v)
    {
        // compute texel coordinates
        // we add 0.5, since the texel is sampled in the middle of the pixel (see OpenGL linear interpolation)
        float px = resX * u + 0.5;
        float py = resY * v + 0.5;

        // find sampling positions of neighbour texels
        Vec2f uv[2][2];
        uv[0][0] = Vec2f(floor(px - 1), floor(py - 1));
        uv[0][1] = Vec2f(floor(px - 1), floor(py));
        uv[1][0] = Vec2f(floor(px), floor(py - 1));
        uv[1][1] = Vec2f(floor(px), floor(py));

        // get the four neighbours
        // make use of _GetTexel method to ensure also handling of proper wrap modes
        Vec4f tex[2][2];
        tex[0][0] = _GetTexel((int)uv[0][0].x(), (int)uv[0][0].y());
        tex[0][1] = _GetTexel((int)uv[0][1].x(), (int)uv[0][1].y());
        tex[1][0] = _GetTexel((int)uv[1][0].x(), (int)uv[1][0].y());
        tex[1][1] = _GetTexel((int)uv[1][1].x(), (int)uv[1][1].y());

        // get fractional parts of the texel coordinates
        // based on the fractional part we interpolate the values
        double ix, iy;
        float fx = modf(px, &ix);
        float fy = modf(py, &iy);

        // interpolate texels bilinearly
        Vec4f result = tex[0][0]*(1-fx)*(1-fy) + tex[1][0]*fx*(1-fy) + tex[0][1]*(1-fx)*fy + tex[1][1]*fx*fy;

        // return new value
        return result;
    }

    //! texture size in x direction
    int resX;

    //! texture size in y direction
    int resY;

public:

    //! Create empy texture object
    Texture2D(int resX, int resY) : Texture(), resX(resX), resY(resY)
    {
        mImage = new Image(resX, resY);
    };
    
    //! Create texture and load it from file
    Texture2D(char* filename) : Texture(), mImage(NULL), resX(1), resY(1)
    {
        Image* img = new Image(1,1);
        if (img->LoadImage(filename))
        {
            mImage = img;
            resX = img->getWidth();
            resY = img->getHeight();
        }else
            delete img;
    }
    
    //! Create texture directly from image
    Texture2D(Image* img) : Texture()
    {
        mImage = img;
        resX = img->getWidth();
        resY = img->getHeight();
    }

    //! Release used data
    ~Texture2D()
    {
        if (mImage) delete mImage;
    }

    /**
     * Get texel on certain position.
     * Here the correct wrapping mode will be used.
     * Interpolate texels if required.
     **/
    Vec4f GetTexel(float u, float v)
    {
        // check which interpolation mode is used
        switch (getInterpolationMode())
        {
            // nearest interpolation
            case NEAREST:
                return _GetTexel(int(resX * u), int(resY * v));
                break;

            // linear interpolation
            case LINEAR:
                return _GetSmoothTexel(u,v);
                break;

            default:
                break;
        };
        return Vec4f();
    }

    //! Return width of the texture 
    inline int getWidth() const { return resX ;}

    //! Return height of the texture 
    inline int getHeight() const { return resY ;}

};

#endif
