#ifndef VEC2F_HXX
#define VEC2F_HXX

#include <iostream>

#include "stdafx.h"


/** A 3D vector
 */
class Vec2f 
{
    // our friends
    friend float    Dot(const Vec2f & a, const Vec2f & b);
    friend Vec2f    Product(const Vec2f & a, const Vec2f & b);
    friend Vec2f    Min(const Vec2f a, const Vec2f b);
    friend Vec2f    Max(const Vec2f a, const Vec2f b);
    friend Vec2f    fabs(const Vec2f & a);
    friend float    Length(const Vec2f & w);
    friend void     Normalize(Vec2f & v);
    friend Vec2f    operator*(const Vec2f & w, const Vec2f& f);
    friend Vec2f    operator*(const Vec2f & w, float f);
    friend Vec2f    operator*(float f, const Vec2f & w);
    friend Vec2f    operator/(const Vec2f & w, float f);
    friend Vec2f    operator/(float f, const Vec2f & w);
 
public:
    Vec2f() 
    {
        v[0] = v[1] = 0.0f;
    }

    Vec2f(float x, float y) 
    {
        v[0] = x;
        v[1] = y;
    }
    
    Vec2f(float f)
    {
        v[0] = f;
        v[1] = f;
    }

    //! x coordinate
    inline float & x()
    {
        return v[0];
    }

    //! const x coordinate
    inline float x() const
    {
        return v[0];
    }

    //! y coordinate
    inline float & y()
    {
        return v[1];
    }

    //! const y coordinate
    inline float y() const
    {
        return v[1];
    }

    //! const index operator
    inline float operator[](int i) const
    {
        return v[i];
    } 

    //! index operator
    inline float & operator[](int i)
    {
        return v[i];
    } 

    inline Vec2f & operator=(const Vec2f & b)
    {
        v[0] = b.v[0];
        v[1] = b.v[1];
        return *this;
    }

    inline int MaxDim() const
    {
        return (v[0] > v[1]) ? 0 : 1;
    }

    inline void SetMin(Vec2f & other)
    { 
        v[0] = MIN(v[0], other.v[0]);
        v[1] = MIN(v[1], other.v[1]);
    }
    
    inline void SetMax(Vec2f & other)
    { 
        v[0] = MAX(v[0], other.v[0]);
        v[1] = MAX(v[1], other.v[1]);
    }

    /*! dot product */
    inline float dot(const Vec2f & b) const
    {
        return v[0]*b.v[0] + v[1]*b.v[1];
    }


    /*! absolute value */
    inline Vec2f fabs() const
    {
        return Vec2f(::fabs(v[0]), ::fabs(v[1]));
    }

    inline float length() const
    {
        return sqrt(Dot(*this, *this));
    }

    /*! normalized version of the vector */
    inline Vec2f normal() const
    {
        return *this * (1.0f / length());
    }

    /*! normalize the vector */
    inline void normalize()
    {
        *this *= 1.0f / length();
    }

    /*! unary - */
    inline Vec2f operator-() const
    {
        return Vec2f(-v[0],-v[1]);
    }

    /*! scale and assign */
    inline void operator*=(float f)
    {
        v[0] *= f;
        v[1] *= f;
    }

    /*! component-wise scale and assign */
    inline void operator*=(const Vec2f & w)
    {
        v[0] *= w.v[0];
        v[1] *= w.v[1];
    }

    /*! scale by reciprocal and assign */
    inline void operator/=(float f)
    {
        (*this) *= (1.0f/f);
    }

    /*! add vectors */
    inline Vec2f operator+(const Vec2f & b) const
    {
        return Vec2f(v[0]+b.v[0], v[1]+b.v[1]);
    }

    /*! add and assign */
    inline Vec2f & operator+=(const Vec2f & b)
    {
        v[0] += b.v[0];
        v[1] += b.v[1];
        return *this;
    }

    /*! substract veltors */
    inline Vec2f operator-(const Vec2f & b) const
    {
        return Vec2f(v[0]-b.v[0], v[1]-b.v[1]);
    }

    /*! substract and assign */
    inline Vec2f & operator-=(const Vec2f & b)
    {
        v[0] -= b.v[0];
        v[1] -= b.v[1];
        return *this;
    }

private:
    float v[2];
};


/*! dot product */
inline float Dot(const Vec2f & a, const Vec2f & b)
{
    return a.v[0]*b.v[0] + a.v[1]*b.v[1];
}

/*! component-wise product */
inline Vec2f Product(const Vec2f & a, const Vec2f & b)
{
    return Vec2f(a.v[0]*b.v[0], a.v[1]*b.v[1]);
}

inline Vec2f Min(const Vec2f a, const Vec2f b)
{
    return Vec2f(MIN(a.v[0], b.v[0]), MIN(a.v[1], b.v[1]));
}

inline Vec2f Max(const Vec2f a, const Vec2f b)
{
    return Vec2f(MAX(a.v[0], b.v[0]), MAX(a.v[1], b.v[1]));
}

/*! absolute value */
inline Vec2f fabs(const Vec2f & a)
{
    return Vec2f(fabs(a.v[0]), fabs(a.v[1]));
}

inline float Length(const Vec2f & w)
{
    return sqrt(Dot(w, w));
}

inline void Normalize(Vec2f & v)
{
    v *= (1.f/Length(v));
}

/*! scale vector */
inline Vec2f operator*(const Vec2f & w, float f)
{
    return Vec2f(f*w.v[0], f*w.v[1]);
}

/*! scale vector */
inline Vec2f operator*(const Vec2f & w, const Vec2f& f)
{
    return Vec2f(f.v[0]*w.v[0], f.v[1]*w.v[1]);
}


/*! scale vector */
inline Vec2f operator*(float f, const Vec2f & w)
{
    return Vec2f(f*w.v[0], f*w.v[1]);
}

/*! scale by reciprocal */
inline Vec2f operator/(const Vec2f & w, float f)
{
    return w*(1.0f/f);
}

/*! a strange operation */
inline Vec2f operator/(float f, const Vec2f & w)
{
    return Vec2f(f / w.v[0], f / w.v[1]);
}


#endif

