#ifndef VEC3F_HXX
#define VEC3F_HXX

#include <iostream>

#include "stdafx.h"

/** A 3D vector
 */
class Vec3f 
{
    // our friends
    friend float    Dot(const Vec3f & a, const Vec3f & b);
    friend Vec3f    Cross(const Vec3f & a, const Vec3f & b);
    friend Vec3f    Product(const Vec3f & a, const Vec3f & b);
    friend Vec3f    Min(const Vec3f a, const Vec3f b);
    friend Vec3f    Max(const Vec3f a, const Vec3f b);
    friend Vec3f    fabs(const Vec3f & a);
    friend float    Length(const Vec3f & w);
    friend void     Normalize(Vec3f & v);
    friend Vec3f    operator*(const Vec3f & w, const Vec3f& f);
    friend Vec3f    operator*(const Vec3f & w, float f);
    friend Vec3f    operator*(float f, const Vec3f & w);
    friend Vec3f    operator/(const Vec3f & w, float f);
    friend Vec3f    operator/(float f, const Vec3f & w);
    friend std::ostream & operator<<(std::ostream & o, const Vec3f & w);

public:
    Vec3f() 
    {
        v[0] = v[1] = v[2] = 0.0f;
    }

    Vec3f(float x, float y, float z) 
    {
        v[0] = x;
        v[1] = y;
        v[2] = z;
    }
    
    Vec3f(float f)
    {
        v[0] = f;
        v[1] = f;
        v[2] = f;
    }

    //! x coordinate
    inline float & x()
    {
        return v[0];
    }

    //! const x coordinate
    inline float x() const
    {
        return v[0];
    }

    //! y coordinate
    inline float & y()
    {
        return v[1];
    }

    //! const y coordinate
    inline float y() const
    {
        return v[1];
    }

    //! z coordinate
    inline float & z()
    {
        return v[2];
    }

    //! const z coordinate
    inline float z() const
    {
        return v[2];
    }

    //! const index operator
    inline float operator[](int i) const
    {
        return v[i];
    } 

    //! index operator
    inline float & operator[](int i)
    {
        return v[i];
    } 

    inline Vec3f & operator=(const Vec3f & b)
    {
        v[0] = b.v[0];
        v[1] = b.v[1];
        v[2] = b.v[2];
        return *this;
    }

    inline int MaxDim() const
    {
        return (v[0] > v[1]) ? ((v[0] > v[2]) ? 0 : 2) : ((v[1] > v[2]) ? 1 : 2);
    }

    inline void SetMin(Vec3f & other)
    { 
        v[0] = MIN(v[0], other.v[0]);
        v[1] = MIN(v[1], other.v[1]);
        v[2] = MIN(v[2], other.v[2]);
    }
    
    inline void SetMax(Vec3f & other)
    { 
        v[0] = MAX(v[0], other.v[0]);
        v[1] = MAX(v[1], other.v[1]);
        v[2] = MAX(v[2], other.v[2]);
    }

    /*! dot product */
    inline float dot(const Vec3f & b) const
    {
        return v[0]*b.v[0] + v[1]*b.v[1] + v[2]*b.v[2];
    }

    /*! vector product */
    inline Vec3f cross(const Vec3f & b) const
    {
        return Vec3f(   v[1]*b.v[2] - v[2]*b.v[1],
                        v[2]*b.v[0] - v[0]*b.v[2],
                        v[0]*b.v[1] - v[1]*b.v[0]);
    }

    /*! absolute value */
    inline Vec3f fabs() const
    {
        return Vec3f(::fabs(v[0]), ::fabs(v[1]), ::fabs(v[2]));
    }

    inline float length() const
    {
        return sqrt(Dot(*this, *this));
    }

    /*! normalized version of the vector */
    inline Vec3f normal() const
    {
        return *this * (1.0f / length());
    }

    /*! normalize the vector */
    inline void normalize()
    {
        *this *= 1.0f / length();
    }

    /*! unary - */
    inline Vec3f operator-() const
    {
        return Vec3f(-v[0],-v[1],-v[2]);
    }

    /*! scale and assign */
    inline void operator*=(float f)
    {
        v[0] *= f;
        v[1] *= f;
        v[2] *= f;
    }

    /*! component-wise scale and assign */
    inline void operator*=(const Vec3f & w)
    {
        v[0] *= w.v[0];
        v[1] *= w.v[1];
        v[2] *= w.v[2];
    }

    /*! scale by reciprocal and assign */
    inline void operator/=(float f)
    {
        (*this) *= (1.0f/f);
    }

    /*! add vectors */
    inline Vec3f operator+(const Vec3f & b) const
    {
        return Vec3f(v[0]+b.v[0], v[1]+b.v[1], v[2]+b.v[2]);
    }

    /*! add and assign */
    inline Vec3f & operator+=(const Vec3f & b)
    {
        v[0] += b.v[0];
        v[1] += b.v[1];
        v[2] += b.v[2];
        return *this;
    }

    /*! substract veltors */
    inline Vec3f operator-(const Vec3f & b) const
    {
        return Vec3f(v[0]-b.v[0], v[1]-b.v[1], v[2]-b.v[2]);
    }

    /*! substract and assign */
    inline Vec3f & operator-=(const Vec3f & b)
    {
        v[0] -= b.v[0];
        v[1] -= b.v[1];
        v[2] -= b.v[2];
        return *this;
    }

    /*! cross product */
    inline Vec3f operator^(const Vec3f & b) const
    {
        return Vec3f(   v[1]*b.v[2] - v[2]*b.v[1],
                        v[2]*b.v[0] - v[0]*b.v[2],
                        v[0]*b.v[1] - v[1]*b.v[0]);
    }

private:
    float v[3];
};


/*! dot product */
inline float Dot(const Vec3f & a, const Vec3f & b)
{
    return a.v[0]*b.v[0] + a.v[1]*b.v[1] + a.v[2]*b.v[2];
}

/*! vector product */
inline Vec3f Cross(const Vec3f & a, const Vec3f & b)
{
    return Vec3f(   a.v[1]*b.v[2] - a.v[2]*b.v[1],
                    a.v[2]*b.v[0] - a.v[0]*b.v[2],
                    a.v[0]*b.v[1] - a.v[1]*b.v[0]);
}

/*! component-wise product */
inline Vec3f Product(const Vec3f & a, const Vec3f & b)
{
    return Vec3f(a.v[0]*b.v[0], a.v[1]*b.v[1], a.v[2]*b.v[2]);
}

inline Vec3f Min(const Vec3f a, const Vec3f b)
{
    return Vec3f(MIN(a.v[0], b.v[0]), MIN(a.v[1], b.v[1]), MIN(a.v[2], b.v[2]));
}

inline Vec3f Max(const Vec3f a, const Vec3f b)
{
    return Vec3f(MAX(a.v[0], b.v[0]), MAX(a.v[1], b.v[1]), MAX(a.v[2], b.v[2]));
}

/*! absolute value */
inline Vec3f fabs(const Vec3f & a)
{
    return Vec3f(fabs(a.v[0]), fabs(a.v[1]), fabs(a.v[2]));
}

inline float Length(const Vec3f & w)
{
    return sqrt(Dot(w, w));
}

inline void Normalize(Vec3f & v)
{
    v *= (1.f/Length(v));
}

/*! scale vector */
inline Vec3f operator*(const Vec3f & w, float f)
{
    return Vec3f(f*w.v[0], f*w.v[1], f*w.v[2]);
}

/*! scale vector */
inline Vec3f operator*(const Vec3f & w, const Vec3f& f)
{
    return Vec3f(f.v[0]*w.v[0], f.v[1]*w.v[1], f.v[2]*w.v[2]);
}


/*! scale vector */
inline Vec3f operator*(float f, const Vec3f & w)
{
    return Vec3f(f*w.v[0], f*w.v[1], f*w.v[2]);
}

/*! scale by reciprocal */
inline Vec3f operator/(const Vec3f & w, float f)
{
    return w*(1.0f/f);
}

/*! a strange operation */
inline Vec3f operator/(float f, const Vec3f & w)
{
    return Vec3f(f / w.v[0], f / w.v[1], f / w.v[2]);
}

/*! write to output stream */
inline std::ostream & operator<<(std::ostream & o, const Vec3f & w)
{
    o << "(" << w.v[0] << "," << w.v[1] << "," << w.v[2] << ")";
    return o;
}

#endif

