#ifndef VEC4F_HXX
#define VEC4F_HXX

#include <iostream>

#include "stdafx.h"
#include "Vec3f.hxx"

/** A 3D vector
 */
class Vec4f 
{
    // our friends
    friend float    Dot(const Vec4f & a, const Vec4f & b);
    friend Vec4f    Product(const Vec4f & a, const Vec4f & b);
    friend Vec4f    Min(const Vec4f a, const Vec4f b);
    friend Vec4f    Max(const Vec4f a, const Vec4f b);
    friend Vec4f    fabs(const Vec4f & a);
    friend float    Length(const Vec4f & w);
    friend void     Normalize(Vec4f & v);
    friend Vec4f    operator*(const Vec4f & w, const Vec4f& f);
    friend Vec4f    operator*(const Vec4f & w, float f);
    friend Vec4f    operator*(float f, const Vec4f & w);
    friend Vec4f    operator/(const Vec4f & w, float f);
    friend Vec4f    operator/(float f, const Vec4f & w);

public:
    Vec4f() 
    {
        v[0] = v[1] = v[2] = v[3] = 0.0f;
    }

    Vec4f(const Vec3f& v3)
    {
        v[0] = v3.x();
        v[1] = v3.y();
        v[2] = v3.z();
        v[3] = 0;
    }
    Vec4f(const Vec3f& v3, float w)
    {
        v[0] = v3.x();
        v[1] = v3.y();
        v[2] = v3.z();
        v[3] = w;
    }

    Vec4f(float x, float y, float z, float w) 
    {
        v[0] = x;
        v[1] = y;
        v[2] = z;
        v[3] = w;
    }
    
    Vec4f(float f)
    {
        v[0] = f;
        v[1] = f;
        v[2] = f;
        v[3] = f;
    }

    //! xyz coordinate
    inline Vec3f xyz() const
    {
        return Vec3f(x(), y(), z());
    }

    //! x coordinate
    inline float & x()
    {
        return v[0];
    }

    //! const x coordinate
    inline float x() const
    {
        return v[0];
    }

    //! y coordinate
    inline float & y()
    {
        return v[1];
    }

    //! const y coordinate
    inline float y() const
    {
        return v[1];
    }

    //! z coordinate
    inline float & z()
    {
        return v[2];
    }

    //! const z coordinate
    inline float z() const
    {
        return v[2];
    }

    //! w coordinate
    inline float & w()
    {
        return v[3];
    }

    //! const w coordinate
    inline float w() const
    {
        return v[3];
    }

    //! const index operator
    inline float operator[](int i) const
    {
        return v[i];
    } 

    //! index operator
    inline float & operator[](int i)
    {
        return v[i];
    } 

    inline Vec4f & operator=(const Vec4f & b)
    {
        v[0] = b.v[0];
        v[1] = b.v[1];
        v[2] = b.v[2];
        v[3] = b.v[3];
        return *this;
    }


    inline void SetMin(Vec4f & other)
    { 
        v[0] = MIN(v[0], other.v[0]);
        v[1] = MIN(v[1], other.v[1]);
        v[2] = MIN(v[2], other.v[2]);
        v[3] = MIN(v[3], other.v[3]);
    }
    
    inline void SetMax(Vec4f & other)
    { 
        v[0] = MAX(v[0], other.v[0]);
        v[1] = MAX(v[1], other.v[1]);
        v[2] = MIN(v[2], other.v[2]);
        v[3] = MIN(v[3], other.v[3]);
    }

    /*! dot product */
    inline float dot(const Vec4f & b) const
    {
        return v[0]*b.v[0] + v[1]*b.v[1] + v[2]*b.v[2] + v[3]*b.v[3];
    }

    /*! absolute value */
    inline Vec4f fabs() const
    {
        return Vec4f(::fabs(v[0]), ::fabs(v[1]), ::fabs(v[2]), ::fabs(v[3]));
    }

    inline float length() const
    {
        return sqrt(Dot(*this, *this));
    }

    /*! normalized version of the vector */
    inline Vec4f normal() const
    {
        return *this * (1.0f / length());
    }

    /*! normalize the vector */
    inline void normalize()
    {
        *this *= 1.0f / length();
    }

    /*! unary - */
    inline Vec4f operator-() const
    {
        return Vec4f(-v[0],-v[1],-v[2],-v[3]);
    }

    /*! scale and assign */
    inline void operator*=(float f)
    {
        v[0] *= f;
        v[1] *= f;
        v[2] *= f;
        v[3] *= f;
    }

    /*! component-wise scale and assign */
    inline void operator*=(const Vec4f & w)
    {
        v[0] *= w.v[0];
        v[1] *= w.v[1];
        v[2] *= w.v[2];
        v[3] *= w.v[3];
    }

    /*! scale by reciprocal and assign */
    inline void operator/=(float f)
    {
        (*this) *= (1.0f/f);
    }

    /*! add vectors */
    inline Vec4f operator+(const Vec4f & b) const
    {
        return Vec4f(v[0]+b.v[0], v[1]+b.v[1], v[2]+b.v[2], v[3]+b.v[3]);
    }

    /*! add and assign */
    inline Vec4f & operator+=(const Vec4f & b)
    {
        v[0] += b.v[0];
        v[1] += b.v[1];
        v[2] += b.v[2];
        v[3] += b.v[3];
        return *this;
    }

    /*! substract veltors */
    inline Vec4f operator-(const Vec4f & b) const
    {
        return Vec4f(v[0]-b.v[0], v[1]-b.v[1], v[2]-b.v[2],v[3]-b.v[3]);
    }

    /*! substract and assign */
    inline Vec4f & operator-=(const Vec4f & b)
    {
        v[0] -= b.v[0];
        v[1] -= b.v[1];
        v[2] -= b.v[2];
        v[3] -= b.v[3];
        return *this;
    }

private:
    float v[4];
};


/*! dot product */
inline float Dot(const Vec4f & a, const Vec4f & b)
{
    return a.v[0]*b.v[0] + a.v[1]*b.v[1] + a.v[2]*b.v[2] + a.v[3]*b.v[3];
}

/*! component-wise product */
inline Vec4f Product(const Vec4f & a, const Vec4f & b)
{
    return Vec4f(a.v[0]*b.v[0], a.v[1]*b.v[1], a.v[2]*b.v[2], a.v[3]*b.v[3]);
}

inline Vec4f Min(const Vec4f a, const Vec4f b)
{
    return Vec4f(MIN(a.v[0], b.v[0]), MIN(a.v[1], b.v[1]), MIN(a.v[2], b.v[2]), MIN(a.v[3], b.v[3]));
}

inline Vec4f Max(const Vec4f a, const Vec4f b)
{
    return Vec4f(MAX(a.v[0], b.v[0]), MAX(a.v[1], b.v[1]), MAX(a.v[2], b.v[2]), MAX(a.v[3], b.v[3]));
}

/*! absolute value */
inline Vec4f fabs(const Vec4f & a)
{
    return Vec4f(fabs(a.v[0]), fabs(a.v[1]), fabs(a.v[2]), fabs(a.v[3]));
}

inline float Length(const Vec4f & w)
{
    return sqrt(Dot(w, w));
}

inline void Normalize(Vec4f & v)
{
    v *= (1.f/Length(v));
}

/*! scale vector */
inline Vec4f operator*(const Vec4f & w, float f)
{
    return Vec4f(f*w.v[0], f*w.v[1], f*w.v[2], f*w.v[3]);
}

/*! scale vector */
inline Vec4f operator*(const Vec4f & w, const Vec4f& f)
{
    return Vec4f(f.v[0]*w.v[0], f.v[1]*w.v[1], f.v[2]*w.v[2], f.v[3]*w.v[3]);
}


/*! scale vector */
inline Vec4f operator*(float f, const Vec4f & w)
{
    return Vec4f(f*w.v[0], f*w.v[1], f*w.v[2], f*w.v[3]);
}

/*! scale by reciprocal */
inline Vec4f operator/(const Vec4f & w, float f)
{
    return w*(1.0f/f);
}

/*! a strange operation */
inline Vec4f operator/(float f, const Vec4f & w)
{
    return Vec4f(f / w.v[0], f / w.v[1], f / w.v[2], f / w.v[3]);
}


#endif

